/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of libcftp - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License (see COPYING.LGPL 
 * for details).
 *
 * $Id: clib2.c,v 1.6 2000/04/23 23:07:41 malekith Exp $
 */

#include "config.h"
#include "clib_int.h"

#include <stdarg.h>
#include <string.h>

struct pkt_response *get_response(CTX *ctx)
{
	struct packet *p;
	struct pkt_response *r;
	
	if ((p = recv_packet_x(ctx, PKT_RESPONSE)) == 0)
		return 0;

	r = xmalloc(p->size + 1);
	memcpy(r, p->data, p->size);
	((char*)r)[p->size] = 0;
	r->no = n2h(r->no);
	free_packet(p);
	return r;
}

void free_resp(struct pkt_response *r)
{
	//xfree0(r, r->size + sizeof(struct pkt_response) - 1);
	xfree(r);
}

struct pkt_response *get_response_x(CTX * ctx, ... )
{
	struct pkt_response *r;
	va_list ap;
	int allow;

try_again:
	r = get_response(ctx);
	
	if (!r)
		return 0;
		
	va_start(ap, ctx);
	while ((allow = va_arg(ap, int)) != 0)
		if (allow == r->no)
			break;
	va_end(ap);
	
	if (allow)
		return r;
	
	set_error(ctx, xprintf("got unexpected response #%d [%s]", r->no, r->data));
	free_resp(r);
	
	if (r->no&RSP_READ_ON) /* READ_on ones should be read to the end ... */
		goto try_again;
		
	return 0;
}

int send_cmd(CTX *ctx, int cmd, const char *data)
{
	int n;
	struct pkt_command *buf;
	
	if (!data)
		data = "";
		
	n = strlen(data) + 4;
	buf = (struct pkt_command *)xmalloc(n);
	buf->no = h2n(cmd);
	memcpy(buf->data, data, strlen(data));
	if (send_packet(ctx, PKT_COMMAND, (unsigned char *)buf, n))
		return -1;
	return 0;
}

static void my_strncpy(byte *dest, const byte *src, int n)
{
	strncpy(dest, src, n-1);
	dest[n-1] = 0;
}
int clib_ls(CTX * ctx, struct clib_stat *buf)
{
	struct rsp_ls fi;
	byte *r;
	char *modes;
	
	if (!ctx->u->ls_in_progress) {
		if (send_cmd(ctx, CMD_LS, 0))
			return -1;
		ctx->u->ls_in_progress = 1;
		ctx->u->ls_last = 0;
		ctx->u->ls_resp = 0;
	}
	
try_get_more:
	if (!ctx->u->ls_resp) {
		if (ctx->u->ls_last) {
			ctx->u->ls_in_progress = 0;
			buf->mode = -1;
			buf->name[0] = 0;
			return 1;	/* EOF */
		} else
			ctx->u->ls_resp =
			    get_response_x(ctx, RSP_LS,
					   RSP_LS | RSP_READ_ON, 0);
		if (!ctx->u->ls_resp)
			return -1;
		if ((ctx->u->ls_resp->no & RSP_READ_ON) == 0)
			ctx->u->ls_last = 1;
		ctx->u->ls_ptr = ctx->u->ls_resp->data;
	}
	
	memcpy(&fi, ctx->u->ls_ptr, sizeof(fi));
	if (fi.mode == (u32)-1) {
		free_resp(ctx->u->ls_resp);
		ctx->u->ls_resp = 0;
		goto try_get_more;
	}
	
	buf->mode = n2h(fi.mode);
	buf->mtime = n2h(fi.mtime);
	buf->size = n2h(fi.size);
	buf->nlink = n2h(fi.nlink);
	
	r = ctx->u->ls_ptr + 16 /* sizeof(struct rsp_ls) - 1 */ ;

	my_strncpy(buf->user, r, sizeof(buf->user)); while (*r) r++; r++;
	my_strncpy(buf->group, r, sizeof(buf->group)); while (*r) r++; r++;
	my_strncpy(buf->name, r, sizeof(buf->name)); while (*r) r++; r++;
	if (buf->mode & MODE_SYMLINK) {
		my_strncpy(buf->symlink, r, sizeof(buf->symlink));
		while (*r) r++; r++;
	} else
		buf->symlink[0] = 0;
	
	if (buf->mode & MODE_DIRECTORY)
		buf->isdir = 1;
	else
		buf->isdir = 0;
		
	modes = buf->text_mode;
	strcpy(modes, "----------");
	if (buf->mode&MODE_DIRECTORY) modes[0] = 'd';
	if (buf->mode&MODE_SYMLINK) modes[0] = 'l';
	if (buf->mode&0400) modes[1] = 'r';
	if (buf->mode&0200) modes[2] = 'w';
	if ((buf->mode&0100) && (buf->mode&04000)) modes[3] = 's';
	if ((buf->mode&0100) && !(buf->mode&04000)) modes[3] = 'x';
	if (!(buf->mode&0100) && (buf->mode&04000)) modes[3] = 'S';
	if (buf->mode&040) modes[4] = 'r';
	if (buf->mode&020) modes[5] = 'w';
	if ((buf->mode&010) && (buf->mode&02000)) modes[6] = 's';
	if ((buf->mode&010) && !(buf->mode&02000)) modes[6] = 'x';
	if (!(buf->mode&010) && (buf->mode&02000)) modes[6] = 'S';
	if (buf->mode&04) modes[7] = 'r';
	if (buf->mode&02) modes[8] = 'w';
	if ((buf->mode&01) && (buf->mode&01000)) modes[9] = 't';
	if ((buf->mode&01) && !(buf->mode&01000)) modes[9] = 'x';
	if (!(buf->mode&01) && (buf->mode&01000)) modes[9] = 'T';
	
	ctx->u->ls_ptr = r;

	return 0;
}

/* ask server to send file NAME 
 * if (whole) { startpos & nbytes are ignored }
 * if (defalte)  { try to request compression } 
 * if (mode) { *mode = mode of file being sent }
 * if (size) { *size = how much bytes will be sent }
 */
int clib_get(struct clib_state *ctx, 
	     const char *name, bool whole, bool deflate, int startpos, 
	     int nbytes, int *mode, int *size, char **owner, char **group)
{
	struct pkt_response *r;
	struct rsp_get *g;
	struct pkt_command *cmd;
	struct cmd_get *cg;
	int n;
	int err;

	/* n = 4 + sizeof(struct cmd_get) - 1 + strlen(name) + 1; */
	n = 4 + 12 + strlen(name) + 1;
	
	cmd = (struct pkt_command *)xmalloc(n);
	cmd->no = h2n(CMD_GET);
	cg = (struct cmd_get *)cmd->data;
	cg->flags = 0;
	if (whole)
		cg->flags |= GET_WHOLE;
	if (deflate && ctx->can_deflate)
		cg->flags |= GET_DEFLATE;
	cg->flags= h2n(cg->flags);
	cg->start_at = h2n(startpos);
	cg->nbytes = h2n(nbytes);
	strcpy(cg->name, name);
	err = send_packet(ctx, PKT_COMMAND, (byte*)cmd, n);
	xfree(cmd);
	
	if (err)
		return -1;
	r = get_response_x(ctx, RSP_GET, RSP_STATUS_ERR, 0);
	if (!r)
		return -2;
		
	if (r->no == RSP_STATUS_ERR) {
		set_error(ctx, xprintf("GET %s: %s", name, r->data));
		free_resp(r);
		return 1;
	}
	
	g = (struct rsp_get *)r->data;
	if (mode)
		*mode = n2h(g->mode);
	if (size)
		*size = n2h(g->size);
	if (owner)
		*owner = xstrdup(g->data);
	if (group)
		*group = xstrdup(strchr(g->data, 0) + 1);
	free_resp(r);
	
	ctx->u->get_aborting = 0;
	ctx->u->get_in_progress = 1;
	
	return 0;	
}

int clib_read(CTX *ctx, void **buf, int *size)
{
	struct packet *p;
	
	p = recv_packet(ctx);
	if (!p)
		return -1;
	*buf = 0;
	*size = 0;
	
	if (p->id == PKT_EOF) {
		struct pkt_response *r;
		int e;
			
		r = (struct pkt_response *) p->data;
		r->no = n2h(r->no);
		ctx->u->get_aborting = 0;
		ctx->u->get_in_progress = 0;
		
		switch (r->no) {
		case RSP_EOF_ABORTED:
			e = 2;
			break;
		case RSP_EOF_OK:
			e = 1;
			break;
		default:
			e = -1;
			set_error(ctx,xprintf("strange response: %d: '%s'", r->no,
			     r->data));
			break;
		}
		free_packet(p);
		return e;
	} else if (p->id == PKT_FILE) {
		*buf = xmalloc(p->size);
		*size = p->size;
		memcpy(*buf, p->data, p->size);
		free_packet(p);
		return 0;
	} else {
		set_error(ctx, xprintf("unexpected packet : %d", p->id));
		free_packet(p);
		return -1;
	}
}

int clib_abort_get(struct clib_state *ctx)
{
	if (!ctx->u->get_in_progress)
		return -1;
	if (ctx->u->get_aborting)
		return 1;
	if (send_packet(ctx, PKT_ABORT, 0, 0))
		return -2;
	return 0;
}

int clib_stor(CTX *ctx, const char *name, 
              int flags, int mode, int startpos, int nbytes,
	      bool deflate)
{
	struct pkt_response *r;
	struct pkt_command *cmd;
	struct cmd_stor *cs;
	int n,err;
	
	/* n = 4 + sizeof(struct cmd_stor) - 1 + strlen(name) + 1; */
	n = 4 + 16 + strlen(name) + 1;
	
	cmd = (struct pkt_command *)xmalloc(n);
	cmd->no = h2n(CMD_STOR);
	cs = (struct cmd_stor *)cmd->data;
	cs->flags = h2n(flags);
	cs->start_at = h2n(startpos);
	cs->nbytes = h2n(nbytes);
	cs->mode = h2n(mode&07777);
	strcpy(cs->name, name);
	err = send_packet(ctx, PKT_COMMAND, (byte*)cmd, n);
	xfree(cmd);
	
	if (err)
		return -1;
	
	r = get_response_x(ctx, RSP_STATUS_OK, RSP_STATUS_ERR, 0);
	if (!r)
		return -2;
		
	if (r->no == RSP_STATUS_ERR) {
		set_error(ctx, xprintf("STOR %s: %s", name, r->data));
		free_resp(r);
		return 1;
	}
	
	if (ctx->can_deflate == 0)
		deflate = 0;
		
	if (z_start_compress(ctx, PKT_FILE, deflate ? 6 : 0))
		return -3;
	
	return 0;
}

int clib_write(CTX *ctx, const void *buf, int size)
{
	if (z_compress(ctx, buf, size))
		return -1;
	return 0;
}

int clib_stor_finish(CTX *ctx)
{
	struct packet *p;
	struct pkt_response *r;
	int e;

	if (z_finish_compress(ctx))
		return -1;
		
	if (send_packet(ctx, PKT_EOF, 0, 0))
		return -1;
	p = recv_packet_x(ctx, PKT_EOF);
	if (!p)
		return -2;
		
	r = (struct pkt_response *)p->data;
	r->no = n2h(r->no);
	
	switch (r->no) {
	case RSP_EOF_OK:
		e = 0;
		break;
	case RSP_EOF_TOO_FEW:
		e = 1;
		break;
	case RSP_EOF_NO_FILE:
		proto_err(ctx, "RSP_EOF_NO_FILE");
		e = 2;
		break;
	case RSP_EOF_TOO_MANY:
		proto_err(ctx, "RSP_EOF_TOO_MANY");
		e = 3;
		break;
	case RSP_EOF_WRITE_ERROR:
		set_error(ctx, xprintf("server got write error: %s", r->data));
		e = 4;
		break;
	default:
		set_error(ctx, xprintf("unexpected RSP(%d) '%s'", r->no, r->data));
		e = -3;
		break;
	}
	free_packet(p);
	return e;
}

char *clib_info(struct clib_state *ctx)
{
	struct pkt_response *r;
	char *i;
	
	if (send_cmd(ctx, CMD_INFO, 0))
		return 0;
		
	r = get_response_x(ctx, RSP_INFO, 0);
	if (!r) return 0;
	i = xstrdup(r->data);
	free_resp(r);
	return i;
}

char *clib_pwd(struct clib_state *ctx)
{
	struct pkt_response *r;
	char *i = 0;

	if (send_cmd(ctx, CMD_PWD, 0))
		return 0;
		
	r = get_response_x(ctx, RSP_CWD, RSP_INTERNAL, 0);
	if (!r) return 0;
	if (r->no == RSP_INTERNAL)
		set_error(ctx, xprintf("internal server error: %s", r->data));
	else
		i = xstrdup(r->data);
	free_resp(r);
	return i;
}

static int yes_no(CTX *ctx)
{
	struct pkt_response *r;
	r = get_response_x(ctx, RSP_STATUS_OK, RSP_STATUS_ERR);
	if (!r)
		return -2;
	if (r->no == RSP_STATUS_OK)
		return 0;
	set_error(ctx, xprintf("error: %s", r->data));
	return -1;
}

int clib_cd(struct clib_state *ctx, const char *arg)
{
	if (send_cmd(ctx, CMD_CD, arg))
		return -1;
	return yes_no(ctx);
}

int clib_rm(struct clib_state *ctx, const char *arg)
{
	if (send_cmd(ctx, CMD_RM, arg))
		return -1;
	return yes_no(ctx);
}

int clib_mkdir(struct clib_state *ctx, const char *arg)
{
	if (send_cmd(ctx, CMD_MKDIR, arg))
		return -1;
	return yes_no(ctx);
}

int clib_rmdir(struct clib_state *ctx, const char *arg)
{
	if (send_cmd(ctx, CMD_RMDIR, arg))
		return -1;
	return yes_no(ctx);
}

int clib_rmr(struct clib_state *ctx, const char *arg)
{
	if (send_cmd(ctx, CMD_RMR, arg))
		return -1;
	return yes_no(ctx);
}
