/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of libcftp - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License (see COPYING.LGPL 
 * for details).
 *
 * $Id: rsa.c,v 1.2 2000/04/23 16:39:03 malekith Exp $
 */

#include "config.h"
#include "gmp.h"
#include "rsa.h"
#include "utils.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

int rsa_verbose = 0;

void rsa_generate_key(struct rsa_key *key, int bits)
{
	mpz_t p,q,pq1,f;
	mpz_t tmp;
	
	mpz_init(p);
	mpz_init(q);
	mpz_init(pq1);
	mpz_init(key->n);
	mpz_init(key->e);
	mpz_init(key->d);
	mpz_init(tmp);
	mpz_init(f);
	
	key->bits = bits;
	
	if (rsa_verbose)
		fprintf(stderr, "generating p: ");
		
	/* 1024 bits key means that pq is 1024 bits long */
	xmpz_get_prime(p, bits / 2);
	
	if (rsa_verbose)
		fprintf(stderr, "\ngenerating q: ");
	xmpz_get_prime(q, bits / 2);
	
	if (rsa_verbose)
		fprintf(stderr, "\ncomupting key .... ");
	
	mpz_mul(key->n, p, q);
	mpz_sub_ui(p, p, 1); 
	mpz_sub_ui(q, q, 1); 
	mpz_mul(pq1, p, q);
	
	mpz_gcd(f,p,q);
	mpz_div(f,pq1,f);
	
	mpz_clear(p);
	mpz_clear(q);
	
	mpz_set_ui(key->e, 0x1f);
	
	do {
		mpz_add_ui(key->e, key->e, 2);
		mpz_gcd(tmp, key->e, pq1);
	} while (mpz_cmp_ui(tmp, 1) != 0);
	
	mpz_invert(key->d, key->e, f);
	mpz_mod(key->d, key->d, f);
	
	mpz_gcd(tmp, key->d, key->n);
	if (mpz_cmp_ui(tmp, 1) != 0)
		abort();
	
	mpz_clear(tmp);
	mpz_clear(pq1);
	
	if (rsa_verbose)
		fprintf(stderr, "ok\n");
}

void rsa_destroy_key(struct rsa_key *key)
{
	mpz_clear(key->n);
	mpz_clear(key->e);
	mpz_clear(key->d);
	key->bits = 0;
}

void rsa_encrypt(mpz_t out, mpz_t in, struct rsa_key *key)
{
	mpz_powm(out, in, key->e, key->n);
}

void rsa_decrypt(mpz_t out, mpz_t in, struct rsa_key *key)
{
	mpz_powm(out, in, key->d, key->n);
}

static void *xxrealloc(void *p, size_t o, size_t n)
{
	void *r = xmalloc(n);
	memcpy(r, p, o < n ? o : n);
	xfree0(p, o);
	return r;
}

void rsa_set_memory_alloc(void)
{
	mp_set_memory_functions(xmalloc, xxrealloc, xfree0);
}

int rsa_read_pub_key(struct rsa_key *pub, const char *data)
{
	mpz_init(pub->n);
	mpz_init(pub->e);
	mpz_init(pub->d);
	pub->bits = -1;

	if (strncmp("cftp public key version 1.0", data, 27) != 0)
		return -1;
	data += 28;
	while (isspace(*data))
		data++;
	if (!strchr(data, '\n'))
		return -1;
	*(strchr(data, '\n')) = 0;
	
	pub->bits = atoi(data);
	data = strchr(data, 0) + 1;
	if (!strchr(data, '\n'))
		return -1;
	*(strchr(data, '\n')) = 0;
	
	mpz_set_str(pub->n, data, 16);
	data = strchr(data, 0) + 1;
	if (!strchr(data, '\n'))
		return -1;
	*(strchr(data, '\n')) = 0;
	
	mpz_set_str(pub->e, data, 16);
	return 0;
}
