/*
 * Copyright (c) y2k Micha Moskal <malekith@topnet.pl>
 *
 * This file is part of cryptoFTP - cftp protocol implementation.
 * CryptoFTP is free software. You can redistribute it and/or modify it under
 * the terms of the GNU General Public License (see COPYING.GPL for details).
 *
 * $Id: server.c,v 1.7 2000/05/05 22:33:10 malekith Exp $
 */

#include "config.h"
#include "server.h"
#include "version.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

struct clib_state *ctx;

const char *auth_user(char buf[], int buflen);

extern char *user_home, *user_name, *user_shell;
extern int user_id, user_gid, shell_type;

/* note that some PKT_MSG must be sent ! -- it marks successful auth
 * to client */
static void send_motd()
{
	struct stat st;
	
	if (stat(MOTD_NAME, &st) == 0 && st.st_size < MOTD_LIMIT) {
		char *buf;
		FILE *f;
	
		buf = xmalloc(st.st_size);
		f = fopen(MOTD_NAME, "r");
		if (!f) die("couldn't open motd");
		fread(buf, 1, st.st_size, f);
		fclose(f);
		send_packet(ctx, PKT_MSG, buf, st.st_size);
		xfree(buf);
	} else
		send_packet(ctx, PKT_MSG, "welcome to cryptoFTP server !", 29);
}

static void do_auth(struct packet *p)
{
	const char *r;
	r = auth_user(p->data, p->size);
	if (r) {
		notice("user %s failed to login because: '%s'", user_name, r);
		send_packet(ctx, PKT_ACCESS_DENIED, (byte*)r, strlen(r));
		sleep(1);
		exit(0);
	}
	setgid(user_gid);
	setuid(user_id);

	notice("user %s logged in", user_name);
	
	if (chdir(user_home))
		chdir("/");
	
	send_motd();
}

void string_resp(int resp, const char *data)
{
	int n;
	struct pkt_response *buf;
	
	n = strlen(data) + 4;
	buf = xmalloc(n);
	buf->no = h2n(resp);
	memcpy(buf->data, data, strlen(data));
	send_packet(ctx, PKT_RESPONSE, (byte *)buf, n);
}

void eof_resp(int resp, const char *data)
{
	int n;
	struct pkt_response *buf;
	
	n = strlen(data) + 4;
	buf = xmalloc(n);
	buf->no = h2n(resp);
	memcpy(buf->data, data, strlen(data));
	send_packet(ctx, PKT_EOF, (byte *)buf, n);
}

void resp(int r)
{
	r = h2n(r);
	send_packet(ctx, PKT_RESPONSE, (byte *)&r, 4);
}

char *get_group(int gid);
char *get_user(int uid);

struct ls_entry {
	struct ls_entry *next;
	char *name;
};

static struct ls_entry *ls_readdir(void)
{
	DIR *d;
	struct dirent *e;
	struct ls_entry *head = 0, *lsp;
	
	d = opendir(".");
	if (!d) {
		string_resp(RSP_INTERNAL, "can't open ``.''");
		notice("ls: cannot open ``.''");
		return 0;
	}
	while ((e = readdir(d))) {
		struct ls_entry *pp = (struct ls_entry *)xmalloc(sizeof(struct ls_entry));
		pp->next=0;
		pp->name=xstrdup(e->d_name);
		if (!head)
			head = pp;
		else if (strcmp(head->name, pp->name) > 0) 
			{ pp->next = head; head = pp; }
		else {
			for (lsp = head; lsp && lsp->next; lsp = lsp->next)
				if (strcmp(lsp->next->name, e->d_name) > 0)
					break;
			pp->next = lsp->next;
			lsp->next = pp;
		}
	}
	closedir(d);
	return head;
}


static void ls_send(struct ls_entry *head)
{
	char buf[CHUNK_SIZE];
	struct ls_entry *lsp;
	int i;
	
	*((int*)buf) = h2n(RSP_LS|RSP_READ_ON);
	i = 4;
	while (head) {
		struct rsp_ls fi;
		struct stat st;
		char *user, *group;
		int size;
		char link_buf[1024];
		
		if (lstat(head->name, &st) < 0) {
			/* this can happen if dir has r-- and we're (for
			 * some reason) inside, perhaps after changing
			 * permisions. */
			enotice("ls: can't stat '%s'", head->name);
                        /* lets trick ... */
                        st.st_gid = 0;
                        st.st_uid = 0;
                        st.st_mode = 0;
                        st.st_mtime = 0;
                        st.st_size = 0;
                        st.st_nlink = 0;
		}
		group = get_group(st.st_gid);
		user = get_user(st.st_uid);
		/* size = sizeof(struct rsp_ls) - 1*/ 
		size = 16 + strlen(user) + 1 + strlen(group) + 1 + 
			strlen(head->name) + 1; 
		link_buf[0] = 0;
		if (S_ISLNK(st.st_mode)) {
			int n;
			n = readlink(head->name, link_buf, sizeof(link_buf) - 1);
			if (n == -1)
				link_buf[0] = 0;
			else
				link_buf[n] = 0;
		}
		if (link_buf[0])
			size += strlen(link_buf) + 1;
		
		if (i + size > (int)sizeof(buf) - 10) {
			*((int*)(buf+i)) = h2n(-1);
			send_packet(ctx, PKT_RESPONSE, buf, i + 4);
			i = 4;
		}
		
		if (i + size > (int)sizeof(buf) - 10) {
			string_resp(RSP_INTERNAL|RSP_READ_ON, 
				"file name too long, didn't fit 4k buffer");
			continue;
		}
		
		fi.mtime = h2n(st.st_mtime);
		fi.size = h2n(st.st_size);
		fi.nlink = h2n(st.st_nlink);
		fi.mode = st.st_mode & 07777;
		if (S_ISDIR(st.st_mode))
			fi.mode |= MODE_DIRECTORY;
		if (link_buf[0])
			fi.mode |= MODE_SYMLINK;
		fi.mode = h2n(fi.mode);
		
		memcpy(buf+i, &fi, 16);
		i += 16;
		
		strcpy(buf + i, user);
		i += strlen(user) + 1;
		
		strcpy(buf + i, group);
		i += strlen(group) + 1;
		
		strcpy(buf + i, head->name);
		i += strlen(head->name) + 1;

		if (link_buf[0]) {
			strcpy(buf + i, link_buf);
			i += strlen(link_buf) + 1;
		}
		lsp = head;
		head = head->next;
		xfree(lsp->name);
		xfree(lsp);
	}
	/* and send the last one */
	*((int*)buf) = h2n(RSP_LS);
	*((int*)(buf+i)) = h2n(-1);
	send_packet(ctx, PKT_RESPONSE, buf, i + 4);
}

static void send_ls()
{
	struct ls_entry *head;
	head = ls_readdir();
	if (head)
		ls_send(head);
}

static void sys_resp(int res)
{
	if (res == -1)
		string_resp(RSP_STATUS_ERR, errno ? strerror(errno) : 
							"unknown error");
	else
		resp(RSP_STATUS_OK);
}

static void send_cwd()
{
#ifdef __linux__
	char *cwd;
	cwd = getcwd(0, 0);
	if (cwd) {
		string_resp(RSP_CWD, cwd);
		free(cwd);
	} else
		string_resp(RSP_INTERNAL, "cannot get cwd");
#else /* !__linux__ */
	char buf[2000];
	if (getcwd(buf, sizeof(buf)))
		string_resp(RSP_CWD, buf);
	else
		string_resp(RSP_INTERNAL, "patch name too long");
#endif /* __linux__ */
}

static int do_rmr(const char *name, int lev)
{
	DIR *d;
	struct dirent *e;
	struct stat buf;
	int r = 0;
	
	/* this shouldn't happen ... */
	if (lev > 50) {
		string_resp(RSP_STATUS_ERR, "dir struct too deep");
		return -1;
	}
	
	if (chdir(name) || (d = opendir("."))==0) {
		sys_resp(-1);
		return -1;
	}
	
	while ((e = readdir(d))) {
		if (strcmp(e->d_name, ".") == 0 || strcmp(e->d_name, "..") == 0)
			continue;
		if (lstat(e->d_name, &buf))
			goto shit;
		if (S_ISLNK(buf.st_mode) || !S_ISDIR(buf.st_mode)) {
			if (unlink(e->d_name))
				goto shit;
		} else {
			if (do_rmr(e->d_name, lev+1))
				goto shit2;
			if (rmdir(e->d_name))
			    	goto shit;
		}
		continue;
	shit:
		sys_resp(-1);
	shit2:
		r = -1;
		break;
	}
	closedir(d);
	chdir("..");
	return r;
}

static void rmr(const char *name)
{
	struct stat buf;
	char cwd[2000];
	
	if (lstat(name, &buf) < 0) {
		sys_resp(-1);
		return;
	}
	if (S_ISLNK(buf.st_mode) || !S_ISDIR(buf.st_mode)) {
		sys_resp(unlink(name));
		return;
	}
	
	if (getcwd(cwd, sizeof(cwd)) == 0)
		sys_resp(-1);
	else {
		int r;
		
		r = do_rmr(name, 0);
		chdir(cwd);
		if (r == 0)
			sys_resp(rmdir(name));
	}
}

int stor_fd = -1, get_fd = -1;
int stor_bytes, get_bytes;

#define STOR_NO_FILE -1
#define STOR_STATUS_OK -2
#define STOR_TOO_MANY_PACKETS -3
#define STOR_WRITE_ERROR -4

static void xstor(byte *data)
{
	struct cmd_stor *si;
	int oflags;
	
	if (stor_fd > 0) {
		close(stor_fd);
		stor_fd = STOR_NO_FILE;
	}
	
	si = (struct cmd_stor *)data;
	si->flags = n2h(si->flags);
	si->start_at = n2h(si->start_at);
	si->nbytes = n2h(si->nbytes);
	si->mode = n2h(si->mode) & 07777;
	
	oflags = O_WRONLY | O_CREAT | O_BINARY;
	if (si->flags & STOR_TRUNC)
		oflags |= O_TRUNC;
	if (si->flags & STOR_EXCL)
		oflags |= O_EXCL;
	if (si->flags & STOR_APPEND) {
		si->start_at = 0;
		oflags |= O_APPEND;
	}
	stor_bytes = si->nbytes;
	info("cmd: STOR '%s' @ %d, %d bytes, mode 0%o", si->name, 
		si->start_at, si->nbytes, si->mode);
	sys_resp(stor_fd = open(si->name, oflags, si->mode));
	if (stor_fd != -1 && si->start_at)
		lseek(stor_fd, si->start_at, SEEK_SET);
	if (stor_fd == -1 && errno == EACCES)
		warn("PERMISSION DENIED to STOR '%s'", si->name);
	else if (stor_fd == -1)
		einfo("STOR '%s' failed", si->name);
}

static void send_rsp_get(int fd, int size)
{
	struct stat st;
	struct pkt_response *rsp;
	struct rsp_get *r;
	char *u = "", *g = "";
	int n;
	int mode = 0644;

	if (fstat(fd, &st) == 0) {
		u = get_user(st.st_uid);
		g = get_group(st.st_gid);
		if (size == -1)
			size = st.st_size;
		mode = st.st_mode & 07777;
	}
	n = 4 + sizeof(struct rsp_get) - 1 + strlen(u) + 1 + strlen(g) + 1;
	rsp = (struct pkt_response *) xmalloc(n);
	rsp->no = h2n(RSP_GET);
	r = (struct rsp_get *) rsp->data;
	r->mode = h2n(mode);
	r->size = h2n(size);
	strcpy(r->data, u);
	strcpy(strchr(r->data, 0) + 1, g);

	send_packet(ctx, PKT_RESPONSE, (byte*)rsp, n);
	xfree(rsp);
}

static void xget(byte * data)
{
	struct cmd_get *gi;
	int compress = 0;

	if (get_fd != -1) {
		close(get_fd);
		get_fd = -1;
	}

	gi = (struct cmd_get *) data;
	gi->flags = n2h(gi->flags);
	gi->start_at = n2h(gi->start_at);
	gi->nbytes = n2h(gi->nbytes);

	get_bytes = gi->nbytes;

	if (gi->flags & GET_WHOLE)
		get_bytes = -1;
	if (gi->flags & GET_DEFLATE)
		compress = 6;
		
	info("cmd: GET '%s' @ %d, %d bytes", gi->name, gi->start_at, get_bytes);
	
	if ((get_fd = open(gi->name, O_RDONLY | O_BINARY)) != -1 &&
	    lseek(get_fd, gi->start_at, SEEK_SET) == -1) {
		int e = errno;
		close(get_fd);
		errno = e;
		get_fd = -1;
	}

	if (get_fd == -1) {
		/* log luseres trying to get /etc/shadow */
		if (errno == EACCES)
			warn("PERMISSION DENIED to GET '%s'", gi->name);
		else
			einfo("GET '%s' failed", gi->name);
		sys_resp(-1);
		return;
	} else
		send_rsp_get(get_fd, get_bytes);

	if (get_bytes == 0) {
		eof_resp(RSP_EOF_OK, "NULL transfer");
		close(get_fd);
		get_fd = -1;
		return;
	}
	z_start_compress(ctx, PKT_FILE, compress);
	
	for (;;) {
		char buf[CHUNK_SIZE];
		int n = CHUNK_SIZE, k;

		if (get_bytes != -1 && n > get_bytes)
			n = get_bytes;k=n;
		n = read(get_fd, buf, n);
		if (n <= 0)
			break;

		z_compress(ctx, buf, n);
		if (get_bytes != -1 && (get_bytes -= n) == 0)
			break;
		if (check_in(ctx)) {
			struct packet *p;
			p = recv_packet(ctx);
			if (p->id != PKT_ABORT) {
				eof_resp(RSP_PROTOCOL_VIOLATION, 
					"got something else then PKT_ABORT "
					"during GET. aborting");
				die("non PKT_ABORT during GET");
			}
			close(get_fd);
			get_fd = -1;
			free_packet(p);
			eof_resp(RSP_EOF_ABORTED, "GET aborted");
			return;
		}
	}
	z_finish_compress(ctx);
	
	eof_resp(RSP_EOF_OK, "");
	close(get_fd);
	get_fd = -1;
}

static void got_file_pkt(byte *data, int size)
{
	int n,r,i;
	
	if (stor_fd == STOR_STATUS_OK)
		stor_fd = STOR_TOO_MANY_PACKETS;
	if (stor_fd < 0)
		return; /* drop packet on the floor */
	i = 0;
	for (n = size > stor_bytes ? stor_bytes : size; n; n -= r, i += r)
		if ((r = write(stor_fd, data + i, n)) == -1) {
			close(stor_fd);
			stor_fd = STOR_WRITE_ERROR;
			return;
		}
	
	if ((stor_bytes -= size) <= 0) {
		close(stor_fd);
		stor_fd = (stor_bytes == 0) ? 
			STOR_STATUS_OK : STOR_TOO_MANY_PACKETS;
	}
}

static void got_eof_pkt()
{
	switch (stor_fd) {
	case STOR_NO_FILE:
		eof_resp(RSP_EOF_NO_FILE,
			"there is no file to store into");
		break;
	case STOR_STATUS_OK:
		eof_resp(RSP_EOF_OK, "");
		break;
	case STOR_TOO_MANY_PACKETS:
		eof_resp(RSP_EOF_TOO_MANY,
			"recived too many bytes to store");
		break;
	case STOR_WRITE_ERROR:
		eof_resp(RSP_EOF_WRITE_ERROR,
			"write error to file occured");
		break;
	default:
		eof_resp(RSP_EOF_TOO_FEW, 
			"not all packets recived");
		close(stor_fd);
		stor_fd = STOR_NO_FILE;
		break;
	}
}

static char *expand_dir(char *orig)
{
	static char buf[2000];
	if (*orig == '~') {
		strncpy(buf, user_home, sizeof(buf));
		buf[sizeof(buf) - 2] = 0;
		strcat(buf, "/");
		strncpy(buf + strlen(buf), orig + 1,
			sizeof(buf) - strlen(buf) - 3);
		return buf;
	} else
		return orig;
}

void xforward(struct pkt_command *c);

char *commands[] = { COMMAND_NAMES };

void process_command(byte *data, int size)
{
	struct pkt_command *cmd;
	int args_len;
	
	cmd = (struct pkt_command *)data;
	cmd->no = n2h(cmd->no);
	args_len = size - 4;
	
	switch (cmd->no) {
	case CMD_INFO:
	case CMD_PWD:
	case CMD_LS:
		info("cmd: %s", commands[cmd->no]); break;
	case CMD_CD:
	case CMD_MKDIR:
	case CMD_RMDIR:
	case CMD_RM:
	case CMD_RMR:
		info("cmd: %s '%s'", commands[cmd->no], cmd->data); break;
	case CMD_GET:
	case CMD_STOR:
	case CMD_FORWARD:
		/* info("cmd: %s XXX", commands[cmd->no]); */
		break;
	}
	
#define NO_ARGS \
	if (args_len) \
		string_resp(RSP_TOO_MUCH_ARGS | RSP_READ_ON, \
			"too much args - this command requires none")

	errno = 0;
	switch (cmd->no) {
	case CMD_INFO:
		string_resp(RSP_INFO, 
SOFTWARE_VERSION "\n"
"cFTP protocol developed by Michal Moskal <malekith@topnet.pl>\n"
"cryptoFTP client and server are FREE software, released under GNU GPL\n"
"you can obtain a copy of free *nix client and server from\n"
"<http://malekith.topnet.pl>. windoze and AmigaOS clients are commin' up!");
		break;
	case CMD_PWD:
		NO_ARGS;
		send_cwd();
		break;
	case CMD_LS:
		send_ls();
		break;
	case CMD_CD:
		sys_resp(chdir(expand_dir(cmd->data)));
		break;
	case CMD_MKDIR:
		sys_resp(mkdir(cmd->data, 0777));
		break;
	case CMD_RMDIR:
		sys_resp(rmdir(cmd->data));
		break;
	case CMD_RM:
		sys_resp(unlink(cmd->data));
		break;
	case CMD_GET:
		xget(cmd->data);
		break;
	case CMD_STOR:
		xstor(cmd->data);
		break;
	case CMD_RMR:
		rmr(cmd->data);
		break;
	default:
		string_resp(RSP_NO_SUCH_COMMAND, 
			    "sorry, i didn't understand your command");
		info("unknown command: %d",cmd->no);
		break;
	}
}

void key_rsa();

static void bye_bye()
{
	clib_free(ctx);
}

/* in case of normal inetd UNIX server, this is real main. otherwise
 * this function is called by real main. standalone server is
 * meant only for testing !!!! don't use it in normal installations
 */
extern char *conf_file;
void ident_user(int);
int inetd_main(int argc, char **argv, int sock)
{
	if (argc > 1)
		conf_file = argv[1];
	
	ctx = clib_alloc();
	ctx->sock = sock;
	
	atexit(bye_bye);
	
	conf_read();
	log_init();
	info("server startup: config=%s", conf_file);
	
#ifndef BRAINDEAD
	ident_user(sock);
#endif
	
	if (version_agreement(ctx))
		die(ctx->error);
	key_rsa();

	for (;;) {
		struct packet *p;
		p = recv_packet(ctx);
		if (!p) {
			info("server shutdown: no packet [%s]", ctx->error);
			return 0;
		}
		switch (p->id) {
		case PKT_DISCONNECT:
			info("server shutdown: pkt_disconnect");
			return 0;
		case PKT_AUTH:
			do_auth(p);
			break;
		case PKT_COOKIE:
			debug("sending back cookie");
			send_packet(ctx, p->id, p->data, p->size);
			break;
		case PKT_COMMAND:
			if (h2n(*(int*)p->data) == CMD_FORWARD) {
				xforward((struct pkt_command*)p->data);
				break;
			}	
		default:
			if (user_name == 0)
				die("got invalid NO_USERNAME packet (%d)",
				    p->id);
			switch (p->id) {
			case PKT_ERROR:
				die("got PKT_ERROR '%s'", p->size ? 
							(char*)p->data : "");
			case PKT_IGNORE:
				/* yhhm... ignore .... */
				break;
			case PKT_COMMAND:
				process_command(p->data, p->size);
				break;
			case PKT_EOF:
				got_eof_pkt();
				break;
			case PKT_FILE:
				got_file_pkt(p->data, p->size);
				break;
			case PKT_ABORT:
				/* ignore it here */
				break;
			default:
				die("unknown pkt_id == %d", p->id);
			}
		}
		free_packet(p);
	}
}
